﻿/*
Teljesítmény-hangolás (Postgres)

1. rekordok fizikai tárolása a lemezen (tarolas.doc)

Minden tábla és minden index külön fájlba kerül a C:\Program Files\PostgresPlus\8.4SS\data\global
könyvtárban.

Az adattárolás hierarchikus, sok cache-szinttel:

1.AB. szerver
2.op. rendszer
3.diszk-vezérl? (pl. RAID)
4.maga a merevlemez

Az adatbiztonság (durability) szempontjából fontos a write-through 
cache-beállítás (ne write-back legyen!), különben a write-ahead log (WAL) ellenére is 
adatvesztés léphet fel pl. áramkimaradáskor.

2.postgresql.conf fájlban a wal_sync_method beállítása

3.Ha a diszkvezérl?nek van saját eleme, akkor ott elfogadható a write-back.

4.Lemez szintjén: "Írási gyorsítótárazás engedélyezése" "Enable write caching on the disk"
opció ki-ikszelése.

Op. rendszer cache demonstrálása:

*/

select count(*) from northwind.orders;
--másodszorra már rövidebb id? alatt fut le
--ezen az sem változtat, ha újar indítjuk a postgres szolgáltatást

/*
2. tesztkörnyezet készítése
*/
--készítsünk teszt táblákat 10 000 rekorddal

--drop table test;

CREATE TABLE test
(
  id serial NOT NULL,
  id2 integer NOT NULL,
  nev character varying(200) NOT NULL,
  CONSTRAINT test_pkey PRIMARY KEY (id) --ez létrehoz egy btree indexet
)

CREATE OR REPLACE FUNCTION load_test(var_limit integer)
  RETURNS integer AS
$BODY$ 
declare var_i integer;
begin
	FOR var_i IN 1..var_limit LOOP
		--select nextval('test_id_seq') into var_id;
		--raise notice 'id: %', var_id;
		insert into test (id2, nev) values 
			(1,
			cast(repeat(chr(10 + cast(100.0*random() as integer)), 200) as varchar(200)));
	END LOOP;
	return var_i;
end;
$BODY$
  LANGUAGE 'plpgsql';

select load_test(10000);
select count(*) from test;

--drop table test_random;
create table test_random as select * from test order by random();
select * from test_random limit 10; 
	--véletlen sorrend az id-ben

update test_random set id2=id;
--delete from test;
drop table test;

CREATE TABLE test --azért drop/create, hogy a seq, tehát az id mez? újra induljon
(
  id serial NOT NULL,
  id2 integer NOT NULL,
  nev character varying(200) NOT NULL,
  CONSTRAINT test_pkey PRIMARY KEY (id) --ez létrehoz egy btree indexet
)

insert into test (id2, nev) select id2, nev from test_random;

select * from test limit 10; 
	--véletlen sorrend az id2-ben, de növekv? sorrend id-ben: 10000..20000

/*
NE tegyük! (ezután a tábla 2x akkora helyet foglal el)
--tegyük helyre az id sorszámait
select max(id), min(id) from test;
update test set id = id-10000;
*/

--statisztikák újrakészítése
vacuum analyze test; --vacuum most nem kell, mert üres volt a tábla
select * from pg_stats where tablename='test';

--		feltöltés UTÁN hozzuk létre az indexeket
--drop index test_index1;
--drop index test_index2;
create unique index test_index1 on test using btree (id2) ;
create index test_index2 on test using btree (nev) ;

/*
3. nézzük meg, hogy mekkora helyet foglalnak el a tábláink
*/

--1 page = 8 KB
--test tábla indexeinek a mérete (blokk egységekben):
SELECT relname, relkind, reltuples, relpages FROM pg_class 
	WHERE relname like 'test%' and relname not like 'test_random%';

/*
--figyeljük meg, hogy a test_pkey index mérete sokkal nagyobb, mint a test_index1
--a sok insert/delete miatt (57 > 30)

--készítsük újra ezt is:
alter table test drop constraint test_pkey;
--drop index test_pkey; --már nincs!
alter table test add constraint test_pkey PRIMARY KEY (id);

--most már ugyanakkorák: 30 blokk
ez csak akkor látszik, ha nem drop/create, hanem delete test
*/

--az index2 a nagyméret? kulcsok  miatt nagyobb
--ezután újrakészítjük a statisztikákat



/*
4. EXPLAIN: az optimizer,  végrehajtási tervek vizsgálata
*/

explain select * from test;
--"Seq Scan on test  (cost=0.00..404.00 rows=10000 width=212)"

--cost: lemez IO+feldolgozás, a startup cost 0
-- 404 = 304 blokk * 1 + 10 000 sor * cpu_tuple_cost (a cpu_tuple_cost default értéke 0.01)
--hálózati késleltetésekkel nem foglalkozik
--rows: a VÁRHATÓAN majd megvizsgált rekordok száma

--sz?rés a rendezetlen mez?re #1
explain select * from test where id2 < 7000;
--"Seq Scan on test  (cost=0.00..429.00 rows=6998 width=212)"
--"  Filter: (id2 < 7000)"

--mivel a sz?r?feltétel alapján sok rekordot vár, ezért végig akar menni az egész táblán
--nem használ indexet, hanem csak sz?r!
--a költséget a sz?rés növelte: 15 = 10000 * cpu_operator_cost

explain select * from test where id < 7000;
--"Index Scan using test_pkey on test  (cost=0.00..422.75 rows=7000 width=212)"
--"  Index Cond: (id < 7000)"

--mi a különbség?
--az, hogy az id mez? értékei fizikailag növekv? sorrendben vannak
--ezért használ indexet
--az index használata csak kis költségcsökkenést okoz, mert alacsony a szelektivitás (sok a sor)

--sz?rés a rendezetlen mez?re #2
explain select * from test where id2 < 100;
--"Bitmap Heap Scan on test  (cost=5.03..213.05 rows=100 width=212)"
--"  Recheck Cond: (id2 < 100)"
--"  ->  Bitmap Index Scan on test_index1  (cost=0.00..5.00 rows=100 width=0)"
--"        Index Cond: (id2 < 100)"

--csak 100 sort vár a statisztika alapján
--el?ször az index1 alapján megkeresi a megfelel? rekordok címeit, (költség: 5)
--aztán a szül? csomópont sorbarendezi ?ket a fizikai blokkcímek szerint (bitmap heap scan)
--végül kiszedi ?ket a táblából
--a rendezés miatt egy lap beolvasásával több megfelel? rekordot is beszed
--tehát a lemez IO kevesebb lesz, mint a megfelel? rekordok száma

--sz?rés a rendezetlen mez?re #3
explain  analyze select * from test where id2 < 2;
--explain  analyze select * from test where id2 < 3;
--"Index Scan using test_index1 on test  (cost=0.00..8.27 rows=1 width=212)"
--"  Index Cond: (id2 < 2)"

--it már olyan kevés sort vár, hogy nem vesz?dik a 2 lépéses kereséssel
--az indexb?l egyenként továbbmegy minden megfelel? rekord címére, 
--és visszaadja a rekordot

--sz?rés a rendezetlen mez?re #3 plusz egy feltétel
explain  analyze select * from test where id2 < 2 and nev like 'W%';
--"Index Scan using test_index1 on test  (cost=0.00..8.27 rows=1 width=212) (actual time=0.024..0.024 rows=0 loops=1)"
--"  Index Cond: (id2 < 2)"
--"  Filter: ((nev)::text ~~ 'N%'::text)"

--a nev mez?re nincs index, ezért ezt csak az index által behozott rekord sz?résére lehet használni
--a költség nem csökkent

explain  analyze select * from test where id2 < 100 and id > 9000;
--"Bitmap Heap Scan on test  (cost=25.01..59.72 rows=10 width=212) (actual time=0.239..0.268 rows=8 loops=1)"
--"  Recheck Cond: ((id2 < 100) AND (id > 9000))"
--"  ->  BitmapAnd  (cost=25.01..25.01 rows=10 width=0) (actual time=0.226..0.226 rows=0 loops=1)"
--"        ->  Bitmap Index Scan on test_index1  (cost=0.00..5.00 rows=100 width=0) (actual time=0.037..0.037 rows=99 loops=1)"
--"              Index Cond: (id2 < 100)"
--"        ->  Bitmap Index Scan on test_pkey  (cost=0.00..19.75 rows=1000 width=0) (actual time=0.167..0.167 rows=1000 loops=1)"
--"              Index Cond: (id > 9000)"

--a költség kisebb, mint a plusz feltétel nélkül, mert kevesebb sorra számít
--a két index szerint kivett címek halmazának veszi a metszetét, aztán rendez és el?vesz

--de lehet, hogy más esetben érdemesebb az egyik feltételt sz?r?ként kezelni:
explain  analyze select * from test where id2 < 7 and id > 9000;
--"Bitmap Heap Scan on test  (cost=4.30..25.86 rows=1 width=212) (actual time=0.049..0.049 rows=0 loops=1)"
--"  Recheck Cond: (id2 < 7)"
--"  Filter: (id > 9000)"
--"  ->  Bitmap Index Scan on test_index1  (cost=0.00..4.30 rows=6 width=0) (actual time=0.026..0.026 rows=6 loops=1)"
--"        Index Cond: (id2 < 7)"


explain select * from test t inner join test_random tr on t.id2=tr.id where t.id <100
--"Hash Join  (cost=14.25..759.75 rows=100 width=424)"
--"  Hash Cond: (tr.id = t.id2)"
--"  ->  Seq Scan on test_random tr  (cost=0.00..707.00 rows=10000 width=212)"
--"  ->  Hash  (cost=13.00..13.00 rows=100 width=212)"
--"        ->  Index Scan using test_pkey on test t  (cost=0.00..13.00 rows=100 width=212)"
--"              Index Cond: (id < 100)"

--el?ször behozza a test_pkey szerint el?vett 100 rekordot a memóriába és a join feltétel szerint hash-eli
--utána az egész test_random táblát is hash-eli, mert nincs hozzá index
--HASH JOIN: 
--kell egy hash fv. valahány kosárral
--a fv. argumentuma a join alapja
--kiszámítható, nagyjából mekkora lesz egy-egy kosár a hash után (teljes rekordokat tesz bele)
--hash készítés közben folyamatosan írja az összes kosarat lemezre
--utána beolvas egy kosarat (be kell férjen a memóriába) és elvégzi a benne lév? rekordokra
--a kért m?veletet. Aztán jön a köv. kosár stb. 
--(ha csak egy táblát hash-el, lehet csoportosításra és ismétl?dés-sz?résre is használni ezt)

create index test_index3 on test_random using btree (id) ;
analyze test_random; --különben nem vesz tudomást az új indexr?l

explain select * from test t inner join test_random tr on t.id2=tr.id where t.id <100
--"Nested Loop  (cost=0.00..533.12 rows=100 width=424)"
--"  ->  Index Scan using test_pkey on test t  (cost=0.00..13.00 rows=100 width=212)"
--"        Index Cond: (id < 100)"
--"  ->  Index Scan using test_index3 on test_random tr  (cost=0.00..5.19 rows=1 width=212)"
--"        Index Cond: (tr.id = t.id2)"

--nested loop: 
--1. kiválasztja a test_pkey alapján a 100 id-t,
--2. ezek mindegyikét a test_index3 alapján történ? kiválasztás paraméterének használja
--533 = 100 sor * 5.19 + CPU id?

set enable_nestloop = off;  -- ezzel a költség 759 lenne!
explain select * from test t inner join test_random tr on t.id2=tr.id where t.id <100
set enable_nestloop = on; 

explain select * from test t, test_random tr where t.id2=tr.id and t.nev < 'K'

/*
5. Az indextípusok áttekintése

B-tree:

	B-tree ppt: + egyéni feladat
	5-ödrend? fa csomópontjában 2..4 kulcs, 3..5 mutató (a gyökér lehet 1 kulcsos is)**
	beszúrás: levélbe, ha ott nincs hely, kettévágni, és a középs? kulcsot felvinni
	törlés: levélb?l törölhet? 
		vagy gyerek csp-tól felhozható
		vagy a közvetlen szomszédtól kölcsönözhet?
		ha mindkét szomszéd minimális: összeolvasztani egy szomszéddal és fentr?l lehozni egy kulcsot
		(rekurzívan, amíg a ** feltételek nem teljesülnek)
	FELADAT: az utolsó ábrán az U kulcs törlése

	* egyenl?ség és range (<, >) lekérdezésre is használható
	* mivel rendezi a kulcsokat (NULLok a végén)
	* ezért egyb?l rendezett kimenetet tud adni
	* több mez?s is lehet, egy ilyen index használható egymez?s feltétel
		keresésére is, lehet?leg a vezet?/els? mez? legyen ez
	* log komplexitású keresés, gyors beszúrás és törlés
	* OLTP alkalmazásokra, ahol
		o kevés rekordot várunk (különben seq scan)
		o a keresési feltétel el?re látható, indexelhet?
		o és nem túl komplikált 
		o viszont fontos a gyors beszúrás és törlés
	* a B-tree rendje (M): egy teljes kiépítettség? csomópont férjen el egy lemez-blokkban
	  M*kulcsméret + (M-1)*mutató-méret < blokkméret byte-ban

	fill factor: a B-tree leveleit ilyen %-ig tölti fel az els? felépítéskor
	default 90, statikus táblánál 100-ra állítandó

Hash:
	a kosár tartalma a bele jutó rekordcímek egy láncolt listában (vagy sorban egymás után a blokkban)
	rekord törlésekor az egész érintett kosarat végig kell nézni, és az illeszked? rekordot törölni
	a legtöbb kosár egy blokkból álljon (de lehet itt is túlcsordulási blokk)
	ekkor egy pont-keresés csak 1 db lemez IO
	
	* csak egyenl?ség lekérdezésre
	* nem támogatja az IS NULL lekérdezéseket
	* nem ad rendezett kimenetet
	* el?ny: egyrekordos (point) lekérdezéseknél el?nyösebb, mint a B-tree

bittérkép (bitmap):
	* a mez? minden el?forduló értékéhez egy vektor
	* adattárház-alkalmazásokhoz (DSS): jól kombinálható ad hoc pont és range lekérdezésekhez
	* nehézkes a karbantartása, inkább statikus adatokhoz való
	* tömöríthet?ek (run length kód), kicsomagolás nélkül felhasználhatóak
	* Postgres nem támogatja (Oracle, MSSQL igen)

több index kombinálása: bittérkép (annyi vektor, ahány keresési feltétel (index), és annyi elem, ahány sor)
ezeket gyors AND/OR m?veletekkel lehet kombinálni (fenti BitmapAnd m?velet)



INDEXVÁLASZTÁS

Ökölszabályok indextervezéshez
0. alapelv: ismerni és elemezni kell az alkalmazást
1. a legszelektívebb attribútumot kell indexelni
2. a tábla legfontosabb indexét kell klaszterezni (fizikai rekordsorrend)
3. ha range lekérdezés várható, akkor B-tree, ha point akkor hash
4. csak akkor kell az index, ha a keresési id? jobban csökken, mint amennyit a karbantartási id? n?
5. kis táblát nem indexelünk (az egész befér a memóriába)

Önálló feladat: 
===============
Írjunk egy scriptet, amely betesz 600 ezer rekordot a fenti táblába.

Feladatok: 
	* pontszer? és tartomány- lekérdezések válaszidejének 
	  és lekérdezési tervének elemzése index nélkül, btree és hash indexekkel

	* az indexek ára: véletlenszer? OLTP módosítások szimulálása és futtatása index nélkül és vele


Hajrá!
*/

drop table test;
CREATE TABLE test
(
  id serial NOT NULL,
  id2 integer NOT NULL,
  nev character varying(200) NOT NULL
 )

CREATE OR REPLACE FUNCTION load_test(var_limit integer) RETURNS integer AS
$BODY$ 
declare var_i integer;
begin
	FOR var_i IN 1..var_limit LOOP
		--select nextval('test_id_seq') into var_id;
		--raise notice 'id: %', var_id;
		insert into test (id2, nev) values 
			(1, cast(repeat(chr(10 + cast(100.0*random() as integer)), 200) as varchar(200)));
	END LOOP;
	return var_i;
end;
$BODY$
LANGUAGE 'plpgsql';

--delete from test; --17 s
select load_test(600000); --31 s 
select * from test limit 100;
drop table test_random;
create table test_random as select * from test order by random(); --22 s
	--véletlen sorrend az id-ben
update test_random set id2=id; --44 s
drop table test;
CREATE TABLE test
(
  id serial NOT NULL,
  id2 integer NOT NULL,
  nev character varying(200) NOT NULL
 )
insert into test (id2, nev) select id2, nev from test_random; --48 s

--INDEX NÉLKÜL

--pont-lekérdezés
select * from test where id2=31124 --9.5 s --index nélkül seq scan
--ismételt futtatásra 0.3 s körül áll be az op.rendszer gyorsítótárazás miatt
analyze test;
explain analyze select * from test where id2=601124; --költség: 22045

--tartomány-lekérdezés
select * from test where id2 between 450000 and 460000 --2.1 s --index nélkül seq scan
explain analyze select * from test where id2 between 450000 and 460000 --költség: 27182
--kisebb tartomány
select * from test where id2 between 345890 and 345900 --297 ms --index nélkül seq scan
explain analyze select * from test where id2 between 345890 and 345900 --költség: 27182
--igazából nincs különbség az id?ben, mivel az egész táblát végignézi (actual time ugyanannyi: 0.3 s)
--csak a klienssel való kommunikálás tart sokáig az 1. esetben
						
--BTREE index az id2-re
--drop index test_index1;
create index test_index1 on test using btree (id2) ; --1.7 s
analyze test;

--pont-lekérdezés
select * from test where id2=534873 --0 ms
explain analyze select * from test where id2=534873 --költség: 8.33, actual: 0.028

--tartomány-lekérdezés
select * from test where id2 between 450000 and 460000 --2.1 s 
explain analyze select * from test where id2 between 450000 and 460000 --költség: 16329, actual: 137
--a becsült költség ugyan kisebb, mint az index nélkül (ezért használza az indexet), de a tényleges futásid?ben
--nincs javulás, mert a kommunikáció dominál

--kisebb tartomány
select * from test where id2 between 345890 and 345900 --0 ms --index nélkül seq scan
explain analyze select * from test where id2 between 345890 and 345900 --költség: 43, actual: 0.122 
--itt már lényges a különbség, mert relatíve kisebb a kommunikáció súlya

--HASH index 
--lesz-e javulás?

drop index test_index1
create index test_index1 on test using hash (id2) ; --1.5 s

--pont-lekérdezés
select * from test where id2=464873 --0 ms 
--az index elkészítése után közvetlenül 16 ms (utána az index a cache-ben van)
explain analyze select * from test where id2=464873 --költség: 8.35, actual: 0.06 
--nem gyorsabb, mint a btree, mert még kicsi a tábla

--a range lekérdezések nem használják a hash indexet:
explain analyze select * from test where id2 between 345890 and 345900 --seq scan

--ne feledjük, hogy nem számíthatunk arra, hogy a kérdéses tábla a gyorsítótárban van!!



--ADATMÓDOSÍTÁS

--explain analyze select random()

--update
drop index test_index1;
explain analyze update test set nev='változott' where id2=86400;--282 ms költség:25682, actual 262
create index test_index1 on test using btree (id2) ; --1.5 s
explain analyze update test set nev='változott' where id2=52344;--15 ms költség:8.32, actual 0.1

--pozitív az index hatása, mert a keresés gyorsabb, emellett eltörpül az index-karbantartás!

set enable_seqscan = off;
analyze test;
explain analyze update test set nev='változott' where id2=cast(600000.0*random() as integer);


--insert
insert into test (id2, nev) values (333588, 'új érték'); --31 ms
explain analyze insert into test (id2, nev) values (399889, 'új érték'); --költség: 0.01, actual: 0.028
create index test_index2 on test using btree (id) ; --1.5 s
create index test_index3 on test using btree (nev) ; --ez az index nagyon nagy, mert hosszú a mez?: 66 s

insert into test (id2, nev) values (33188, 'új érték'); --219 ms (els? futtatásra) kés?bb csökken, mert az indexek a cache-be kerülnek
explain analyze insert into test (id2, nev) values (39989, 'núj névérték'); 

--több index esetén az insert lassul 

drop table test;
CREATE TABLE test
(
  id serial NOT NULL,
  id2 integer NOT NULL,
  nev character varying(200) NOT NULL
 )
--index nélkül a betöltés 48 s volt.
--insert into test (id2, nev) select id2, nev from test_random; --48 s

--DE ha van már a táblán index az insert indításakor
create index test_index2 on test using btree (id) ; 
create index test_index1 on test using btree (id2) ;
insert into test (id2, nev) select id2, nev from test_random; --61 s
--TEHÁT: nagy mennyiség? adat betöltése el?tt az indexeket törölni kell 
--(a futásid? jelent?s része származik a serial olvasásából)

--ugyanez a delete esetén:
--törlés 2 index-szel
delete from test; --20 s
--index nélkül
drop table test;
CREATE TABLE test
(
  id serial NOT NULL,
  id2 integer NOT NULL,
  nev character varying(200) NOT NULL
)
insert into test (id2, nev) select id2, nev from test_random; --44 s
analyze test;
delete from test; --20 s

--változó terhelés
drop table test;
CREATE TABLE test
(
  id serial NOT NULL,
  id2 integer NOT NULL,
  nev character varying(200) NOT NULL
)
insert into test (id2, nev) select id2, nev from test_random; --44 s
analyze test;

--a script véletlenszer?en végez egy m?veletet
--ezzel próbáljuk szimulálni a valós alkalmazást

CREATE OR REPLACE FUNCTION update_test(var_limit integer, var_update integer, 
var_insert integer, var_delete integer) RETURNS integer AS
$BODY$ 
declare var_i integer;
declare var_id integer;
declare var_dummy integer;
declare var_action integer;
begin
	FOR var_i IN 1..var_limit LOOP
		--select nextval('test_id_seq') into var_id;
		--raise notice 'id: %', var_id;
		var_action := cast(100.0*random() as integer);
		if var_action < var_update then --update
--			raise notice 'update';
			var_id := cast(600000.0*random() as integer);
			update test set nev='változott' where id2 = var_id;		
		elsif ....

--ÖNÁLLÓ FELADAT:
--a többi 3 ágat írjuk meg!!

		end if;
	END LOOP;
	return 1;
end;
$BODY$
LANGUAGE 'plpgsql';

--INDEX NÉLKÜL (futtassuk többször is!)
select * from update_test(100, 30, 50, 70); --19 s
--valószín?ségek: update 30%, insert 20%, delete 20%, select 30%
--(futtassuk más valószín?ségekkel is!)

--INDEXekkel
create index test_index2 on test using btree (id) ; 
create index test_index1 on test using btree (id2) ;
create index test_index3 on test using btree (nev) ;

select * from update_test(100, 30, 50, 70); --14 s, 8s (másodszorra)
--ezért használunk indexeket

--tehetünk még rá többmez?s indexeket, ezzel n? az adminisztrációs teher

/*
TOVÁBBI érdekes témák:

	* több indexet is tud használni, ha pl. mindegyikre van feltétel (kombinálás bittérképpel)
	* ennek alternatívája a többmez?s index: gyorsabb, de az els? mez?re kell legyen feltétel 
	* a legtöbbet használt index legyen klaszterezett: különösen range lekérdezéseknél gyors

ESETTANULMÁNY a többmez?s indexre: TB-adatbázis eseménysorai (távoli SQL-serverr?l)

TELJESÍTMÉNY-HANGOLÁS EGYÉB MÓDSZEREI

1. A hardver

	* RAM, HDD, CPU: sok memória, gyors és hibat?r? diszk-rendszer, RAID 1+0 (lásd ppt), több proc. 
	  Általában az I/O a sz?k keresztmetszet.
	

2. A szerver lényeges beállításai (Postgres)

	* max_connections = <num> — This option sets the maximum number of database backend to have at any one time.
	  Use this feature to ensure that you do not launch so many backends that you begin swapping to disk 
	  and kill the performance of all the children.
	
	* effective_cache_size = <num> — This value tells PostgreSQL's optimizer how much memory PostgreSQL 
	  has available for caching data and helps in determining whether or not it use an index or not. 
	  The larger value increases the likely hood of using an index.

	* shared_buffers = <num> — Editing this option is the simplest way to improve the performance of your 
	  database server. The default is pretty low for most modern hardware. General wisdom says that this 
	  should be set to roughly 25% of available RAM on the system.

3. Adatbázis-tervezés

Az adatmodell követheti az alkalmazás igényeit. Például:

CREATE TABLE employees ( 
    id                int4 PRIMARY KEY,
    active            boolean, 
    first_name        varchar, 
    middle_name       varchar, 
    last_name         varchar, 
    ssn               varchar, 
    address1          varchar, 
    address2          varchar, 
    city              varchar, 
    state             varchar(2), 
    zip               varchar, 
    home_phone        varchar, 
    work_phone        varchar, 
    cell_phone        varchar, 
    fax_phone         varchar, 
    pager_number      varchar, 
    business_email    varchar, 
    personal_email    varchar, 
    salary            int4, 
    vacation_days     int2, 
    sick_days         int2, 
    employee_number   int4, 
    office_addr_1     varchar,
    office_addr_2     varchar, 
    office_city       varchar, 
    office_state      varchar(2), 
    office_zip        varchar, 
    department        varchar,
    title             varchar,
    supervisor_id     int4
); 

ehelyett:

CREATE TABLE employees ( 
    id               int4 PRIMARY KEY, 
    active           boolean, 
    employee_number  int4, 
    first_name       varchar, 
    middle_name      varchar, 
    last_name        varchar, 
    department       varchar, 
    title            varchar, 
    email            varchar
); 

CREATE TABLE employee_address ( 
    id               int4 PRIMARY KEY, 
    employee_id      int4,
    personal         boolean, 
    address_1        varchar, 
    address_2        varchar, 
    city             varchar,
    state            varchar(2),
    zip              varchar
); 

CREATE TABLE employee_number_type (
    id               int4 PRIMARY KEY,
    type             varchar
); 

CREATE TABLE employee_number ( 
    id               int4 PRIMARY KEY,
    employee_id      int4, 
    type_id          int4, 
    number           varchar
); 

CREATE TABLE employee_hr_info ( 
    id               int4 PRIMARY KEY,
    employee_id      int4, 
    ssn              varchar, 
    salary           int4, 
    vacation_days    int2,
    sick_days        int2
); 
 
--így az eredmény gyorsabb m?ködés lesz.

4. Ha már megvan az alkalmazás, de lassú: profiler, tracer stb. eszközökkel a tényleges forgalmat elemezni,
a kritikus lekérdezéseket gyorsítani

HA MÁR SEMMI SEM SEGÍT: gondolkodni és áttervezni az alkalmazást
(pl. QCS rendelésfrissítési view gyorsítása)

*/  